/* cinvert.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: CINVERT.PAS */

#include "cmplxops.h"
#include "diodemx.h"
#include "cinvert.h"


/**
 * Cinvert - complex matrix inversion
 * @a:_: pointer to input matrix to invert
 * @b: pointer to inverted output matrix
 *
 *
 * Cinvert() computes the inverse of the complex matrix pointed by @a_
 * and returns it in @b. It uses Gauss-Jordan reduction WITHOUT PIVOTING
 * a_ must be nonsingular and have no zero entries on the main diagonal
 * WARNING: this subroutine is designed specifically for use with
 * diodemx.c; it may not work well in other applications 
 */
void Cinvert(cmplx_mtx a_, cmplx_mtx b)
{
  short j, k, i;
  complex one, zero, cmplx, cmplx1;
  cmplx_mtx a;


  memcpy(a, a_, sizeof(cmplx_mtx));

  /* define constants */
  one.x = 1.0;
  one.y = 0.0;
  zero.x = 0.0;
  zero.y = 0.0;

  /* initialize b; complex identity matrix */
  for (j = 0; j <= 8; j++) {
    for (k = 0; k <= 8; k++)
      b[j][k] = zero;
    b[j][j] = one;
  }

  for (i = 0; i <= 8; i++) {	/* row no. */
    for (j = 0; j <= 8; j++) {	/* col. no. */
      /* determine multiplying factor */
      if (j != i) {
	cmplx = CDIV(a[j][i], a[i][i]);
	cmplx = CSUB(zero, cmplx);

	/* multiply row and add to all other rows */
	for (k = 0; k <= 8; k++) {
	  cmplx1 = CMULT(cmplx, a[i][k]);
	  a[j][k] = CADD(cmplx1, a[j][k]);
	  cmplx1 = CMULT(cmplx, b[i][k]);
	  b[j][k] = CADD(cmplx1, b[j][k]);
	}
      }
    }
  }

  /* a is now a diagonal; normalize it */
  /* and do the same to b */
  for (j = 0; j <= 8; j++) {
    for (k = 0; k <= 8; k++)
      b[j][k] = CDIV(b[j][k], a[j][j]);
  }
}
