/* cmplxops.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: CMPLXOPS.PAS */

#include <math.h>
#include "cmplxops.h"


/* complex arithmetic subroutines */

complex CADD(complex a, complex b)
{
  complex c;

  c.x = a.x + b.x;
  c.y = a.y + b.y;

  return c;
}


complex CSUB(complex a, complex b)
{
  complex c;

  c.x = a.x - b.x;
  c.y = a.y - b.y;

  return c;
}


complex CMULT(complex a, complex b)
{
  complex c;
  double x1, x2, x3, x4;

  x1 = a.x * b.x;
  x2 = a.y * b.y;
  x3 = a.x + a.y;
  x4 = b.x + b.y;

  c.x = x1 - x2;
  c.y = x3 * x4 - x1 - x2;

  return c;
}


complex CDIV(complex a, complex b)
{
  complex c;
  double denom;

  denom = b.x * b.x + b.y * b.y;

  c.x = (a.x * b.x + a.y * b.y) / denom;
  c.y = (a.y * b.x - a.x * b.y) / denom;

  return c;
}


complex CTAN(complex x)
{
  complex num, denom, tangent;

  num.x = tan(x.x);
  num.y = tanh(x.y);
  denom.x = 1.0;
  denom.y = -tan(x.x) * tanh(x.y);
  tangent = CDIV(num, denom);

  return tangent;
}


double RAD_TO_DEG(double angle)
{
  angle = 180.0 * angle / M_PI;
  if (angle > 180.0)
    angle -= 360.0;
  if (angle < -180.0)
    angle += 360.0;
  return angle;
}


void COM_TO_POLAR(complex x, double *mag, double *phase)
{
  *mag = sqrt(x.x * x.x + x.y * x.y);
  *phase = atan2(x.y, x.x);
}


void POLAR_TO_COM(double mag, double phase, complex * x)
{
  (*x).x = mag * cos(phase);
  (*x).y = mag * sin(phase);
}


complex CSQRT(complex x)
{
  complex y;
  double mag, phase;

  COM_TO_POLAR(x, &mag, &phase);
  y.x = sqrt(mag) * cos(0.5 * phase);
  y.y = sqrt(mag) * sin(0.5 * phase);

  return y;
}


complex CONJ(complex x)
{
  complex y;

  y.x = x.x;
  y.y = -x.y;

  return y;
}


complex CNEG(complex x)
{
  complex y;

  y.x = -x.x;
  y.y = -x.y;

  return y;
}
