/* diodemx.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: DIODEMX.PAS */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <curses.h>
#include "diodemx.h"
#include "cmplxops.h"
#include "dmxutil.h"
#include "dmxdrv.h"
#include "scifunc.h"
#include "dmxlarge.h"
#include "dmxsmall.h"
#include "noise.h"
#include "im.h"
#include "graphit.h"

#define pi              3.1415926535898


static double vlosource, vbias, IFfreq, conv_loss, rev_loss, MAGain, x, Pstart, Pstop, Pstep, Plo_dBm, vstart, vstop, vstep, input_power, IM2a_power, PIM2a_dBm, IM2b_power, PIM2b_dBm, IM3_power, PIM3_dBm, Pin_dBm, IFfreq2, loss, TSSB, Tout, TSSBactual, Toutactual, TSSBconj, Toutconj;
static short j, inpfreq, outfreq, steps, stepnum, Ypos, Xpos;
static complex Zinsig, Zoutsig, Zinmatch, Zoutmatch, cmplx, inpzemb, outzemb, inputZ, outputZ;
static FILE *infile, *soutfile, *loutfile, *vjijfile, *sweepfile;
static StringType IDword, datafilename, filename, sfilename, lgsgfilename, swpfilename, response;
static char ch1, ch3, chxx, option, yes_or_no, A_or_C;
static cmplx_mtx Yconv, Zconv, JYconv, JZconv;
static boolean valid_ldata, valid_sdata, opened_soutfile, wants_IM, is_OK, wants_conj_data, NoError, lastitem, ResetZ, ReadErr;

/* Global variables */
double I0, dt, temp, eta, phi, cj0, zc, rs, ggamma, delta;
complex one, zero, jimag;
complex zemblo[9];
Zsigtype zembsig;
char infile_NAME[FNSIZE];
char soutfile_NAME[FNSIZE];
char loutfile_NAME[FNSIZE];
char vjijfile_NAME[FNSIZE];
char sweepfile_NAME[FNSIZE];

double pdc, Idc, Vdc, Vdc_desired, LOfreq, pwrdiss, pavail, LOeff, tolerance;
complex Zin;
waveform ij, ic, vj, id, vd, c;
boolean LO_display;

/* tmp string for ncurses addstr */
char s[200];
int ioresult;

void error_exit(int code, char *code_desc)
{
  fprintf(stderr, "error %i: %s\n", code, code_desc);
  endwin();
  exit(-1);
}


void FRAME(void)
{
  short j;
  /* sorry, frame drawing not yet ready... */
  if (false) { 
    attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
    move(0, 1);
    addch(ACS_ULCORNER);
    move(0, 78);
    addch(ACS_URCORNER);
    
    move(0, 2);
    for (j = 3; j <= 78; j++)	/*205 */
      addch(ACS_BSBS);
    
    move(24, 1);
    for (j = 3; j <= 79; j++)	/*196 */
      addch(ACS_HLINE);
    
    for (j = 1; j <= 23; j++) {
      move(j, 1);
      addch(ACS_VLINE);
      move(j, 78);
      addch(ACS_VLINE);
    }
    move(24, 1);
    addch(ACS_LLCORNER);
  move(24, 78);
  addch(ACS_LRCORNER);
  refresh();

  attrset(COLOR_PAIR(COLOR_WHITE));
  }
}


static void ReadFile(boolean * Error)
{
  short j;
  char ctest;

  *Error = false;
  /* Begin file read loop */
  do {
    do {
      ReadStr(&infile, IDword, &lastitem);
      if (lastitem) {
	fscanf(infile, "%*[^\n]");
	getc(infile);
      }
    } while (lastitem);
    CapStr(IDword);
    if (!strcmp(IDword, "DIODE")) {
      fscanf(infile, "%lg%lg%lg%lg%lg%lg%lg%*[^\n]", &I0, &eta, &cj0, &phi, &ggamma, &rs, &temp);
      getc(infile);
    } else if (!strcmp(IDword, "VOLTS")) {
      fscanf(infile, "%lg%lg%lg%*[^\n]", &vlosource, &vbias, &Vdc_desired);
      getc(infile);
    } else if (!strcmp(IDword, "FREQS")) {
      fscanf(infile, "%lg%lg%lg%*[^\n]", &LOfreq, &IFfreq, &IFfreq2);
      getc(infile);
    } else if (!strcmp(IDword, "MISCL")) {
      fscanf(infile, "%lg%lg%*[^\n]", &zc, &tolerance);
      getc(infile);
    } else if (!strcmp(IDword, "ZSIG1")) {
      for (j = 0; j <= 3; j++) {
	fscanf(infile, "%lg%lg", &cmplx.x, &cmplx.y);
	cmplx.y = -cmplx.y;
	/* Use conjugate for all LSB impedances */
	if (fabs(cmplx.x) == 0.0) {
	  cmplx.x = 1.0e-3;
	  ResetZ = true;
	}
	zembsig[j] = cmplx;
      }
      fscanf(infile, "%*[^\n]");
      getc(infile);
    } else if (!strcmp(IDword, "ZSIG2")) {
      for (j = 4; j <= 8; j++) {
	fscanf(infile, "%lg%lg", &cmplx.x, &cmplx.y);
	if (fabs(cmplx.x) == 0.0) {
	  cmplx.x = 1.0e-3;
	  ResetZ = true;
	}
	zembsig[j] = cmplx;
      }
      fscanf(infile, "%*[^\n]");
      getc(infile);
    } else if (!strcmp(IDword, "ZELO1")) {
      for (j = 0; j <= 3; j++) {
	fscanf(infile, "%lg%lg", &cmplx.x, &cmplx.y);
	if (fabs(cmplx.x) == 0.0) {
	  cmplx.x = 1.0e-3;
	  ResetZ = true;
	}
	zemblo[j] = cmplx;
      }
      fscanf(infile, "%*[^\n]");
      getc(infile);
    } else if (!strcmp(IDword, "ZELO2")) {
      for (j = 4; j <= 8; j++) {
	fscanf(infile, "%lg%lg", &cmplx.x, &cmplx.y);
	if (fabs(cmplx.x) == 0.0) {
	  cmplx.x = 1.0e-3;
	  ResetZ = true;
	}
	zemblo[j] = cmplx;
      }
      fscanf(infile, "%*[^\n]");
      getc(infile);
    } else {
      printf("Unrecognized code in data file: %s\n", IDword);
      *Error = true;
      goto _LBailout;
    }

    ctest = getc(infile);
    if (ctest != EOF)
      ungetc(ctest, infile);

  } while (!feof(infile));	/* end of file read loop */
  if (infile != NULL)
    fclose(infile);
  infile = NULL;
  if (ResetZ) {
    attrset(COLOR_PAIR(COLOR_RED));
    move(4, 9);
    addstr(" WARNING:  zero real parts of embedding impedances ");
    move(5, 9);
    addstr(" have been reset to 0.001 ohms to prevent a crash. ");
    move(6, 9);
    addstr(" Strike any key to continue                        ");
    refresh();
    ch1 = getch();
  }
_LBailout:;
}				/*ReadFile */


static void ReadFileName(char *DataFileName)
{
  char *TEMP1;

  clear();
  FRAME();
  bkgd(COLOR_PAIR(COLOR_BLUE));
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  move(0, 0);
  addstr("\n                                  DIODEMX\n");
  addstr("                       Diode Mixer Analysis Program\n\n");
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr("                        C Version I.39, 10/27/2002\n");
  addstr("                                   --*--\n\n\n\n");
  refresh();
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  addstr("  Enter data file name: ");
  attrset(COLOR_PAIR(COLOR_WHITE));
  refresh();

  getnstr(DataFileName, 41);
  TEMP1 = strchr(DataFileName, '\n');
  if (TEMP1 != NULL)
    *TEMP1 = 0;
  RemoveSpaces(DataFileName);
  if (strchr(DataFileName, '.') == NULL) {
    /* Pascal version uses ".TXT" */
    strcat(DataFileName, ".txt");
    /* possible string truncation in assignment */
  }
}


static void PutMenu(char *option)
{
  /* Write menu; use lots of snazzy colors */
  bkgd(COLOR_PAIR(COLOR_BLUE));
  clear();
  FRAME();
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  move(1, 0);
  addstr("                                   DIODEMX\n");
  addstr("                        Diode Mixer Analysis Program\n");
  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
  addstr("\n                        C Version I.39, 10/27/2002\n");
  addstr("                                    --*--\n\n");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  addstr("     Options:\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  R");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("ead large-signal data from a file\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  L");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("arge signal analysis\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  F");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("ile large-signal data\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  S");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("mall signal analysis\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  O");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("pen a file to save small-signal data\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  C");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("lose small-signal data file\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  D");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("isplay LO parameters\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  H");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("armonic power levels\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  G");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("raph large-signal data\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  P");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("LO: LO power sweep\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  V");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("b: Bias voltage sweep\n");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr("                  Q");
  attrset(COLOR_PAIR(COLOR_WHITE));
  addstr("uit\n\n");
  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
  addstr("     Enter desired option: ");
  getyx(stdscr, Ypos, Xpos);
  attrset(COLOR_PAIR(COLOR_WHITE));
  mvaddstr(22, 44, "Data file name: ");
  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
  addstr(datafilename);
  move(Ypos, Xpos);
  attrset(COLOR_PAIR(COLOR_WHITE));
  refresh();
  nodelay(stdscr, FALSE);
  *option = getch();
  nodelay(stdscr, TRUE);
  CAP(option);
}				/*PutMenu */


/**************************************************************/
/***********************  MAIN PROGRAM  ***********************/
/**************************************************************/


int main(int argc, char *argv[])
{
  char *TEMP1;
  char STR2[256];
  short FORLIM;

  sweepfile = NULL;
  vjijfile = NULL;
  loutfile = NULL;
  soutfile = NULL;
  infile = NULL;

  /* curses stuff */
  initscr();
  if (has_colors()) {
    start_color();
    init_pair(COLOR_BLACK, COLOR_BLACK, COLOR_WHITE);
    init_pair(COLOR_GREEN, COLOR_GREEN, COLOR_BLUE);
    init_pair(COLOR_RED, COLOR_WHITE, COLOR_RED);
    init_pair(COLOR_CYAN, COLOR_CYAN, COLOR_BLUE);
    init_pair(COLOR_WHITE, COLOR_WHITE, COLOR_BLUE);
    init_pair(COLOR_MAGENTA, COLOR_WHITE, COLOR_MAGENTA);
    init_pair(COLOR_BLUE, COLOR_BLACK, COLOR_BLUE);
    init_pair(COLOR_YELLOW, COLOR_YELLOW, COLOR_BLUE);
  }

  cbreak();
  echo();

  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);

  /* Set complex coefficients */
  zero.x = 0.0;
  zero.y = 0.0;
  one.x = 1.0;
  one.y = 0.0;
  jimag.x = 0.0;
  jimag.y = 1.0;

  /* initialize constants */
  opened_soutfile = false;
  valid_ldata = false;
  valid_sdata = false;
  LO_display = true;
  NoError = true;
  ResetZ = false;
  bkgd(COLOR_PAIR(COLOR_BLUE));
  clear();

  /* Write header */
  if (argc == 1)
    ReadFileName(datafilename);
  else {
    strcpy(datafilename, argv[1]);
    /* possible string truncation in assignment */
  }
  RemoveSpaces(datafilename);
  if (*datafilename == '\0')
    goto _LFINIS;
  strcpy(infile_NAME, datafilename);
  if (infile != NULL)
    infile = freopen(infile_NAME, "r", infile);
  else
    infile = fopen(infile_NAME, "r");

  if (infile == NULL)		/* file not found */
    goto _LFINIS;

  /* Read the file */
  ReadFile(&ReadErr);
  if (ReadErr)
    goto _LFINIS;

  /* Calculate some essentials: */

  /* large-signal integration time interval */
  dt = 1.0 / (32.0 * LOfreq);

  /* diode I/V exponent */
  delta = 1.1744526e04 / (eta * temp);

  /* set capacitor params to (+) if they aren't */
  ggamma = fabs(ggamma);
  phi = fabs(phi);


  do {
    NoError = true;		/* (Don't you wish it were always like this?) */

    PutMenu(&option);

    /* Determine & execute the desired option */
    if (option == 'R') {
      clear();
      move(5, 0);
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" File name ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("(<cr> to abort): ");
      refresh();
      echo();
      nodelay(stdscr, FALSE);
      getnstr(lgsgfilename, 41);
      noecho();
      nodelay(stdscr, TRUE);
      TEMP1 = strchr(lgsgfilename, '\n');
      if (TEMP1 != NULL)
	*TEMP1 = 0;
      if (*lgsgfilename != '\0') {
	if (strchr(lgsgfilename, '.') == NULL) {
	  /* Pascal version uses ".TXT" */
	  strcat(lgsgfilename, ".txt");
	  /* possible string truncation in assignment */
	}
	strcpy(vjijfile_NAME, lgsgfilename);

	if (vjijfile != NULL)
	  vjijfile = freopen(vjijfile_NAME, "r", vjijfile);
	else 
	  vjijfile = fopen(vjijfile_NAME, "r");

	is_OK = (vjijfile != NULL);
	if (!is_OK) {
	  WINDOW *tmp_win;
	  tmp_win = subwin(stdscr, 3, 41, 11, 24);
	  wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	  wclear(tmp_win);
	  wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	  wmove(tmp_win, 0, 0);
	  waddstr(tmp_win, " File does not exist\n");
	  if (valid_ldata)
	    waddstr(tmp_win, " Previous large-signal data is preserved\n");
	  waddstr(tmp_win, " Strike any key to continue");
	  NoError = false;
	} else {
	  fscanf(vjijfile, "%*[^\n]");	/* skip headings */
	  getc(vjijfile);
	  for (j = 0; j <= 31; j++) {
	    fscanf(vjijfile, "%lg%lg%lg%lg%lg%*[^\n]", &ij[j], &id[j], &vj[j], &vd[j], &c[j]);
	    getc(vjijfile);
	    c[j] *= 1.0e-12;
	  }
	  fscanf(vjijfile, "%*[^\n]");	/* skip Plo heading */
	  getc(vjijfile);
	  fscanf(vjijfile, "%lg%*[^\n]", &pavail);
	  getc(vjijfile);
	  if (vjijfile != NULL)
	    fclose(vjijfile);
	  vjijfile = NULL;
	  valid_ldata = true;
	  valid_sdata = false;
	  LO_prams(vd, id, &Zin, &pwrdiss, &Idc, &Vdc);
	  write_LO_prams();
	}
      }
    } else if (option == 'H') {
      if (valid_ldata)
	Harmonics(vd, id);
      else {
	WINDOW *tmp_win;
	tmp_win = subwin(stdscr, 2, 58, 11, 18);
	wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	wclear(tmp_win);
	wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	wmove(tmp_win, 0, 0);
	waddstr(tmp_win, "No valid large-signal data; run LGSG or read from a file\n");
	waddstr(tmp_win, "Strike any key to continue");
	touchwin(stdscr);
	wrefresh(tmp_win);
	NoError = false;
	delwin(tmp_win);
      }
    } else if (option == 'L') {
      echo();
      nodelay(stdscr, FALSE);
      clear();
      move(4, 0);
      cmplx = zemblo[1];
      Plo_dBm = 10.0 * log10(vlosource * vlosource / (8.0 * cmplx.x)) + 30.0;
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Beginning large-signal analysis\n");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("\n Current values:\n");
      sprintf(s, "            Plo, dBm: %7.3f\n", Plo_dBm);
      addstr(s);
      sprintf(s, "               Vbias: %7.4f\n", vbias);
      addstr(s);
      sprintf(s, "          Diode bias: %7.4f\n\n", Vdc_desired);
      addstr(s);
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Want to change anything? ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("(y/n, <cr> to abort): ");
      refresh();
      yes_or_no = getch();
      if (yes_or_no == '\015')
	goto _LCONTINUE;
      if (yes_or_no == 'y' || yes_or_no == 'Y') {
	do {
	  addstr("\n Enter new Plo, Vbias, Diode bias: ");
	  refresh();
	  ioresult = scanw("%lg%lg%lg", &Plo_dBm, &vbias, &Vdc_desired);
	} while (ioresult != 3);
      }
      pavail = 0.001 * X_tothe_Y(10.0, Plo_dBm / 10.0);
      vlosource = sqrt(8.0 * cmplx.x * pavail);
      LO_display = true;
      do_lgsg(vlosource, &vbias);
      valid_ldata = true;
      valid_sdata = false;
      nodelay(stdscr, TRUE);
      noecho();
    } else if (option == 'D') {
      if (valid_ldata)
	write_LO_prams();
      else {
	WINDOW *tmp_win;
	tmp_win = subwin(stdscr, 2, 58, 11, 18);
	wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	wclear(tmp_win);
	wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	wmove(tmp_win, 0, 0);
	waddstr(tmp_win, "No valid large-signal data; run LGSG or read from a file\n");
	waddstr(tmp_win, "Strike any key to continue");
	touchwin(stdscr);
	wrefresh(tmp_win);
	NoError = false;
	delwin(tmp_win);
      }
    } else if (option == 'F') {
      if (valid_ldata) {
	clear();
	move(4, 0);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" File name ");
	attrset(COLOR_PAIR(COLOR_WHITE));
	addstr("(<cr> to abort): ");
	refresh();
	echo();
	nodelay(stdscr, FALSE);
	getnstr(filename, 41);
	noecho();
	nodelay(stdscr, TRUE);
	TEMP1 = strchr(filename, '\n');
	if (TEMP1 != NULL)
	  *TEMP1 = 0;
	if (*filename != '\0') {
	  if (strchr(filename, '.') == NULL) {
	    /* Pascal version uses ".TXT" */
	    strcat(filename, ".txt");
	    /* possible string truncation in assignment */
	  }
	  strcpy(loutfile_NAME, filename);
	  if (loutfile != NULL)
	    loutfile = freopen(loutfile_NAME, "w", loutfile);
	  else
	    loutfile = fopen(loutfile_NAME, "w");
	  if (loutfile == NULL)
	    error_exit(FileNotFound, loutfile_NAME);
	  fprintf(loutfile, "    ij       id       vj       vd       c (pF)\n");
	  for (j = 0; j <= 31; j++) {
	    x = c[j] * 1.0e12;
	    fprintf(loutfile, "%9.5f%9.5f%9.4f%9.4f%9.5f\n", ij[j], id[j], vj[j], vd[j], x);
	  }
	  fprintf(loutfile, "Plo:\n");
	  fprintf(loutfile, "% .7E\n\n", pavail);
	  fprintf(loutfile, "LO Frequency:%8.4f GHz\n\n", 1.0e-9 * LOfreq);
	  fprintf(loutfile, "Zs:   Real part=%7.2f     Imag part=%7.2f\n", zemblo[1].x, zemblo[1].y);
	  fprintf(loutfile, "Zin:  Real part=%7.2f     Imag part=%7.2f\n", Zin.x, Zin.y);
	  fprintf(loutfile, "Idc, mA=%7.3f       Vdc=%8.3f\n", 1000.0 * Idc, Vdc);
	  fprintf(loutfile, "Diode RF power dissipation, mW=%7.2f\n", 1000.0 * (pwrdiss - pdc));
	  fprintf(loutfile, "Diode DC power dissipation, mW=%7.2f\n", 1000.0 * pdc);
	  fprintf(loutfile, "Available RF LO power, mW=%7.2f     dBm: %5.2f\n", 1000.0 * pavail, 10.0 * log10(pavail) + 30.0);
	  fprintf(loutfile, "LO Efficiency=%7.4f\n", LOeff);
	  fclose(loutfile);
	  loutfile = NULL;
	}
      } else {
	WINDOW *tmp_win;
	tmp_win = subwin(stdscr, 2, 58, 11, 18);
	wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	wclear(tmp_win);
	wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	wmove(tmp_win, 0, 0);
	waddstr(tmp_win, "No valid large-signal data; run LGSG or read from a file\n");
	waddstr(tmp_win, "Strike any key to continue");
	touchwin(stdscr);
	wrefresh(tmp_win);
	NoError = false;
	delwin(tmp_win);
      }
    } else if (option == 'S') {
      if (valid_ldata) {
	clear();
	move(5, 0);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" Begin small-signal analysis: takes about 10 seconds");
	refresh();
	if (!valid_sdata) {
	  Small_sig(c, ij, LOfreq, IFfreq, Yconv, Zconv, JYconv, JZconv);
	  valid_sdata = true;
	}
	do {
	  wants_IM = false;
	  echo();
	  nodelay(stdscr, FALSE);
	  do {
	    clear();
	    move(1, 0);
	    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	    addstr("   Enter input and output frequency indices, n1, n2: ");
	    getyx(stdscr, Ypos, Xpos);
	    addstr("\n\n");
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    addstr("   Note:\n");
	    attrset(COLOR_PAIR(COLOR_WHITE));
	    addstr("      Input freq  =");
	    attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
	    addstr(" n1");
	    attrset(COLOR_PAIR(COLOR_WHITE));
	    addstr("*f(LO) + f(0)\n");
	    addstr("      Output freq =");
	    attrset(COLOR_PAIR(COLOR_MAGENTA) | A_BOLD);
	    addstr(" n2");
	    attrset(COLOR_PAIR(COLOR_WHITE));
	    addstr("*f(LO) + f(0)\n\n");
	    addstr("   and\n");
	    sprintf(s, "       f(0) =%8.4f GHz\n", 1.0e-9 * IFfreq);
	    addstr(s);
	    sprintf(s, "      f(LO) =%8.4f GHz\n", 1.0e-9 * LOfreq);
	    addstr(s);
	    move(Ypos, Xpos);
	    refresh();
	    ioresult = scanw("%hd%hd", &inpfreq, &outfreq);
	    is_OK = (ioresult == 2 && abs(inpfreq) <= 4 && abs(outfreq) <= 4);
	    /* and (inpfreq<>outfreq)) ; */
	  } while (!is_OK);
	  noecho();
	  nodelay(stdscr, TRUE);
	  clear();
	  move(1, 0);
	  addstr("\n");
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Noise analysis; takes about 10 seconds");
	  refresh();
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  SSPrams(inpfreq, outfreq, Yconv, &Zinsig, &Zoutsig, &Zinmatch, &Zoutmatch, &conv_loss, &rev_loss, &MAGain);
	  NOISE(JYconv, zembsig, Zinmatch, Zoutmatch, Zoutsig, inpfreq, outfreq, conv_loss, MAGain, ij, false, &TSSBactual, &Toutactual);
	  NOISE(JYconv, zembsig, Zinmatch, Zoutmatch, Zoutsig, inpfreq, outfreq, conv_loss, MAGain, ij, true, &TSSBconj, &Toutconj);

	  /* Write small-sig results */
	  clear();
	  move(1, 0);
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Small-signal source & load impedances:\n");
	  inpzemb = zembsig[inpfreq + 4];
	  if (inpfreq < 0)
	    inpzemb.y = -inpzemb.y;
	  outzemb = zembsig[outfreq + 4];
	  if (outfreq < 0)
	    outzemb.y = -outzemb.y;
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Source: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "Real part=%7.2f   Imag part=%7.2f;   Freq=%8.4f GHz\n", inpzemb.x, inpzemb.y, 1.0e-9 * fabs(inpfreq * LOfreq + IFfreq));
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Load:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "   Real part=%7.2f   Imag part=%7.2f;   Freq=%8.4f GHz\n\n", outzemb.x, outzemb.y, 1.0e-9 * fabs(outfreq * LOfreq + IFfreq));
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Small-signal performance:\n");
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Zin:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "  Real part=%7.2f     Imag part=%7.2f\n", Zinsig.x, Zinsig.y);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Zout:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, " Real part=%7.2f     Imag part=%7.2f\n", Zoutsig.x, Zoutsig.y);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Conversion loss: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%6.2f dB", conv_loss);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr("  Rev. loss: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%6.2f dB\n", rev_loss);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr("            TSSB:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%7.2f deg. K   ", TSSBactual);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr("Tout: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%6.2f deg. K\n\n", Toutactual);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Conjugate-match performance:\n");
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Zin:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "  Real Part=%7.2f     Imag part=%7.2f\n", Zinmatch.x, Zinmatch.y);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Zout:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, " Real Part=%7.2f     Imag part=%7.2f\n", Zoutmatch.x, Zoutmatch.y);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Conj match loss: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%6.2f dB\n", MAGain);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr(" Conj Match TSSB:");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%7.2f deg. K   ", TSSBconj);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	  addstr("Tout: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  sprintf(s, "%6.2f deg. K\n\n", Toutconj);
	  addstr(s);
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Want IM calculations?");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(" (y/n): ");
	  refresh();
	  echo();
	  nodelay(stdscr, FALSE);
	  yes_or_no = getch();
	  nodelay(stdscr, TRUE);
	  noecho();
	  addstr("   ");
	  if (yes_or_no == 'y' || yes_or_no == 'Y') {
	    wants_IM = true;
	    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	    addstr(" Enter input power, dBm: ");
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    getyx(stdscr, Ypos, Xpos);
	    refresh();
	    echo();
	    nodelay(stdscr, FALSE);
	    do {
	      is_OK = true;
	      ioresult = scanw("%lg", &Pin_dBm);
	      if (ioresult != 1) {
		is_OK = false;
		move(Ypos, Xpos);
		clrtoeol();
	      }
	    } while (!is_OK);
	    noecho();
	    nodelay(stdscr, TRUE);
	    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	    addstr(" Actual or conjugate-match data");
	    attrset(COLOR_PAIR(COLOR_WHITE));
	    addstr(" (a/c): ");
	    refresh();
	    echo();
	    nodelay(stdscr, FALSE);
	    ch1 = getch();
	    nodelay(stdscr, TRUE);
	    noecho();

	    if (ch1 == 'c' || ch1 == 'C')
	      wants_conj_data = true;
	    else
	      wants_conj_data = false;
	    input_power = 0.001 * X_tothe_Y(10.0, Pin_dBm / 10.0);
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    addstr(" Beginning IM analysis; takes about 50 seconds");
	    getyx(stdscr, Ypos, Xpos);
	    refresh();
	    IM(inpfreq, outfreq, Zconv, Yconv, JZconv, JYconv, LOfreq, IFfreq, IFfreq2, input_power, vj, ij, c, zembsig, Zinmatch, Zoutmatch, wants_conj_data, &IM2a_power, &IM2b_power, &IM3_power);
	    PIM2a_dBm = 10.0 * log(IM2a_power) / log(10.0) + 30.0;
	    PIM2b_dBm = 10.0 * log(IM2b_power) / log(10.0) + 30.0;
	    PIM3_dBm = 10.0 * log(IM3_power) / log(10.0) + 30.0;
	    move(Ypos - 1, 0);
	    clrtoeol();
	    move(Ypos, 0);
	    clrtoeol();
	    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	    if (wants_conj_data)
	      addstr(" Conjugate-match IM levels:\n");
	    else
	      addstr(" IM levels with actual source/load impedances:\n");
	    attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	    addstr(" IM output (dBm) w1-w2: ");
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    sprintf(s, "%7.2f", PIM2a_dBm);
	    addstr(s);
	    attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	    addstr("   2w1: ");
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    sprintf(s, "%7.2f", PIM2b_dBm);
	    addstr(s);
	    attrset(COLOR_PAIR(COLOR_CYAN) | A_BOLD);
	    addstr("  2w1-w2: ");
	    attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	    sprintf(s, "%7.2f\n", PIM3_dBm);
	    addstr(s);
	    refresh();
	  }
	  if (opened_soutfile) {
	    attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	    addstr(" Want to save this output? ");
	    attrset(COLOR_PAIR(COLOR_WHITE));
	    addstr("(y/n): ");
	    refresh();
	    echo();
	    nodelay(stdscr, FALSE);
	    yes_or_no = getch();
	    nodelay(stdscr, TRUE);
	    noecho();
	    if (yes_or_no == 'y' || yes_or_no == 'Y') {
	      fprintf(soutfile, "\nSmall-sig results for (%d %d)\n\n", inpfreq, outfreq);
	      fprintf(soutfile, "Plo, mW=%7.2f  Vdc=%7.4f  Idc, mA=%6.4f\n", 1000.0 * pavail, Vdc, 1000.0 * Idc);
	      fprintf(soutfile, "Source impedance: Real part=%6.2f   Imag part=%6.2f;   Freq=%8.4f GHz\n", inpzemb.x, inpzemb.y, 1.0e-9 * fabs(inpfreq * LOfreq + IFfreq));
	      fprintf(soutfile, "Load impedance:   Real part=%6.2f   Imag part=%6.2f;   Freq=%8.4f GHz\n", outzemb.x, outzemb.y, 1.0e-9 * fabs(outfreq * LOfreq + IFfreq));
	      fprintf(soutfile, "Zin:  Real part=%7.2f     Imag part=%7.2f\n", Zinsig.x, Zinsig.y);
	      fprintf(soutfile, "Zout: Real part=%7.2f     Imag part=%7.2f\n", Zoutsig.x, Zoutsig.y);
	      fprintf(soutfile, "Conjugate match impedances:\n");
	      fprintf(soutfile, "Zin:  Real Part=%7.2f     Imag part=%7.2f\n", Zinmatch.x, Zinmatch.y);
	      fprintf(soutfile, "Zout: Real Part=%7.2f     Imag part=%7.2f\n", Zoutmatch.x, Zoutmatch.y);
	      fprintf(soutfile, "Conversion loss: %6.2f dB\n", conv_loss);
	      fprintf(soutfile, "   Reverse loss: %6.2f dB\n", rev_loss);
	      fprintf(soutfile, "Conj match loss: %6.2f dB\n", MAGain);
	      fprintf(soutfile, "           TSSB: %7.2f deg. K     Tout: %6.2f deg. K\n", TSSBactual, Toutactual);
	      fprintf(soutfile, "Conj Match TSSB: %7.2f deg. K     Tout: %6.2f deg. K\n", TSSBconj, Toutconj);
	      if (wants_IM) {
		if (wants_conj_data)
		  fprintf(soutfile, "Conjugate match IM data\n");
		else
		  fprintf(soutfile, "IM data with actual source/load Z\n");
		fprintf(soutfile, "IM output power (dBm), w1-w2: %7.2f   2w1: %7.2f    2w1-w2: %7.2f\n", PIM2a_dBm, PIM2b_dBm, PIM3_dBm);
	      }
	    }
	  }
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Want another small-signal run? ");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr("(y/n): ");
	  refresh();
	  echo();
	  nodelay(stdscr, FALSE);
	  yes_or_no = getch();
	  nodelay(stdscr, TRUE);
	  noecho();
	  CAP(&yes_or_no);
	  move(21, 0);
	  clrtoeol();
	} while (yes_or_no == 'Y');
      } else if (true) {	/* if (!valid_ldata) ??? */
	WINDOW *tmp_win;
	tmp_win = subwin(stdscr, 2, 58, 11, 18);
	wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	wclear(tmp_win);
	wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	wmove(tmp_win, 0, 0);
	waddstr(tmp_win, "No valid large-signal data; run LGSG or read from a file\n");
	waddstr(tmp_win, "Strike any key to continue");
	touchwin(stdscr);
	wrefresh(tmp_win);
	NoError = false;
	delwin(tmp_win);
      }
    } else if (option == 'O') {
      clear();
      echo();
      nodelay(stdscr, FALSE);
      move(5, 0);
      if (opened_soutfile) {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	sprintf(s, " %s is currently open.  Want to close it? ", sfilename);
	addstr(s);
	attrset(COLOR_PAIR(COLOR_WHITE));
	addstr("y/n: ");
	refresh();
	yes_or_no = getch();
	CAP(&yes_or_no);
	addstr("\n");
	refresh();
	if (yes_or_no == 'Y') {
	  if (soutfile != NULL)
	    fclose(soutfile);
	  soutfile = NULL;
	  opened_soutfile = false;
	}
      }
      if (!opened_soutfile) {
	move(7, 0);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" Enter filename for saving small-signal results: ");
	refresh();
	attrset(COLOR_PAIR(COLOR_WHITE));
	getnstr(sfilename, 41);
	TEMP1 = strchr(sfilename, '\n');
	if (TEMP1 != NULL)
	  *TEMP1 = 0;
	if (*sfilename == '\0')
	  strcpy(sfilename, "SMALL");
	if (strchr(sfilename, '.') == NULL) {
	  /* Pascal version uses ".TXT" */
	  strcat(sfilename, ".txt");
	  /* possible string truncation in assignment */
	}
	sprintf(s, "\n File %s opened\n\n", sfilename);
	addstr(s);
	refresh();
	strcpy(soutfile_NAME, sfilename);
	if (soutfile != NULL)
	  soutfile = freopen(soutfile_NAME, "w", soutfile);
	else
	  soutfile = fopen(soutfile_NAME, "w");
	if (soutfile == NULL)
	  error_exit(FileNotFound, soutfile_NAME);
	opened_soutfile = true;
      }
      noecho();
      nodelay(stdscr, TRUE);
    } else if (option == 'C') {
      clear();
      move(5, 0);
      if (opened_soutfile) {
	if (soutfile != NULL)
	  fclose(soutfile);
	soutfile = NULL;
	opened_soutfile = false;
	sprintf(s, "\n Closed %s\n", sfilename);
	addstr(s);
      } else
	addstr(" No data file has been opened\n");
      refresh();
    } else if (option == 'P') {
      echo();
      nodelay(stdscr, FALSE);
      clear();
      move(1, 0);
      LO_display = false;
      wants_IM = false;
      wants_conj_data = false;
      addstr("\n");
      do {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" LO power sweep: enter start, stop, step size (dBm): ");
	refresh();
	attrset(COLOR_PAIR(COLOR_WHITE));
	ioresult = scanw("%lg%lg%lg", &Pstart, &Pstop, &Pstep);
      } while (ioresult != 3);
      Pstep = fabs(Pstep);
      if (Pstart > Pstop)
	Pstep = -Pstep;
      if (Pstep == 0)
	Pstep = 1.0e6;
      addstr("\n");
      do {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" Enter input and output frequency indices; n1, n2: ");
	refresh();
	attrset(COLOR_PAIR(COLOR_WHITE));
	ioresult = scanw("%hd%hd", &inpfreq, &outfreq);
      } while (ioresult != 2);
      addstr("\n");
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Filename to save output data: ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      refresh();
      getnstr(swpfilename, 41);
      TEMP1 = strchr(swpfilename, '\n');
      if (TEMP1 != NULL)
	*TEMP1 = 0;
      if (*swpfilename == '\0')
	strcpy(swpfilename, "PLOSWEEP");
      if (strchr(swpfilename, '.') == NULL) {
	/* Pascal version uses ".TXT" */
	strcat(swpfilename, ".txt");
	/* possible string truncation in assignment */
      }
      addstr("\n");
      sprintf(s, " Data saved in %s\n", swpfilename);
      addstr(s);
      addstr("\n Current values:\n");
      sprintf(s, "   Bias source voltage, Vbias: %7.4f\n", vbias);
      addstr(s);
      sprintf(s, "                   Diode bias: %7.4f\n\n", Vdc_desired);
      addstr(s);
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Want to change anything? ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("(y/n, <cr> to abort): ");
      refresh();
      yes_or_no = getch();
      CAP(&yes_or_no);
      addstr("\n");
      if (yes_or_no == '\015')
	goto _LCONTINUE;
      if (yes_or_no == 'Y') {
	do {
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Enter new Vbias, Diode bias: ");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  refresh();
	  ioresult = scanw("%lg%lg", &vbias, &Vdc_desired);
	} while (ioresult != 2);
      }
      addstr("\n");
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Want IM calculations? ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("y/n: ");
      refresh();
      yes_or_no = getch();
      CAP(&yes_or_no);
      if (yes_or_no == 'Y') {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr("   Enter input power, dBm: ");
	attrset(COLOR_PAIR(COLOR_WHITE));
	refresh();
	getyx(stdscr, Ypos, Xpos);
	do {
	  is_OK = true;
	  ioresult = scanw("%lg", &Pin_dBm);
	  if (ioresult != 1) {
	    is_OK = false;
	    move(Ypos, Xpos);
	    clrtoeol();
	  }
	} while (!is_OK);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	input_power = 0.001 * X_tothe_Y(10.0, Pin_dBm / 10.0);
	wants_IM = true;
      } else {
	addstr("\n");
      }
      addstr("\n");
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Actual or Conjugate-match loss and impedances (a/c): ");
      refresh();
      attrset(COLOR_PAIR(COLOR_WHITE));
      A_or_C = getch();
      noecho();
      nodelay(stdscr, TRUE);

      CAP(&A_or_C);
      addstr("\n");
      if (A_or_C == 'C')
	wants_conj_data = true;
      strcpy(sweepfile_NAME, swpfilename);
      if (sweepfile != NULL)
	sweepfile = freopen(sweepfile_NAME, "w", sweepfile);
      else
	sweepfile = fopen(sweepfile_NAME, "w");
      if (sweepfile == NULL)
	error_exit(FileNotFound, sweepfile_NAME);
      fprintf(sweepfile, "\n LO POWER SWEEP            DATA FILE IS %s\n", datafilename);
      fprintf(sweepfile, " INPUT FREQ:%8.4f GHz   OUTPUT FREQ:%8.4f GHz   LO FREQ:%8.4f GHz\n\n", 1.0e-9 * fabs(inpfreq * LOfreq + IFfreq), 1.0e-9 * fabs(outfreq * LOfreq + IFfreq), 1.0e-9 * LOfreq);
      if (wants_IM)
	fprintf(sweepfile, " RF INPUT POWER= %6.2f dBm\n\n", Pin_dBm);
      if (wants_conj_data)
	fprintf(sweepfile, " Tabulated results are CONJUGATE MATCH values\n");
      else
	fprintf(sweepfile, " Tabulated results are ACTUAL values\n");
      fprintf(sweepfile, "\n  PLO    Vdc    Idc        Zin          Zout        Lc      PIM(dBm)        TSSB\n");
      fprintf(sweepfile, " (dBm)          (mA)    Re     Im     Re     Im    (dB) w1-w2  2w1  2w1-w2 deg. K\n");
      putc(' ', sweepfile);
      for (j = 2; j <= 81; j++)
	putc('=', sweepfile);
      fprintf(sweepfile, "\n\n");
      steps = labs((long) floor((Pstop - Pstart) / Pstep + 0.5));
      FORLIM = steps;
      for (stepnum = 0; stepnum <= FORLIM; stepnum++) {
	Plo_dBm = Pstart + Pstep * stepnum;
	pavail = 0.001 * X_tothe_Y(10.0, Plo_dBm / 10.0);
	cmplx = zemblo[1];
	vlosource = sqrt(8.0 * cmplx.x * pavail);
	do_lgsg(vlosource, &vbias);
	move(7, 36);
	attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	addstr("PERFORMING SMALL-SIG ANALYSIS");
	attrset(COLOR_PAIR(COLOR_WHITE));
	refresh();
	Small_sig(c, ij, LOfreq, IFfreq, Yconv, Zconv, JYconv, JZconv);
	SSPrams(inpfreq, outfreq, Yconv, &Zinsig, &Zoutsig, &Zinmatch, &Zoutmatch, &conv_loss, &rev_loss, &MAGain);
	NOISE(JYconv, zembsig, Zinmatch, Zoutmatch, Zoutsig, inpfreq, outfreq, conv_loss, MAGain, ij, wants_conj_data, &TSSB, &Tout);
	addstr("\n");
	refresh();
	if (wants_IM) {
	  move(8, 36);
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("   PERFORMING IM ANALYSIS");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  refresh();
	  IM(inpfreq, outfreq, Zconv, Yconv, JZconv, JYconv, LOfreq, IFfreq, IFfreq2, input_power, vj, ij, c, zembsig, Zinmatch, Zoutmatch, wants_conj_data, &IM2a_power, &IM2b_power, &IM3_power);
	  PIM2a_dBm = 10.0 * log(IM2a_power) / log(10.0) + 30.0;
	  PIM2b_dBm = 10.0 * log(IM2b_power) / log(10.0) + 30.0;
	  PIM3_dBm = 10.0 * log(IM3_power) / log(10.0) + 30.0;
	} else {
	  PIM2a_dBm = 0.0;
	  PIM2b_dBm = 0.0;
	  PIM3_dBm = 0.0;
	}
	if (wants_conj_data) {
	  loss = MAGain;
	  inputZ = Zinmatch;
	  outputZ = Zoutmatch;
	} else {
	  loss = conv_loss;
	  inputZ = Zinsig;
	  outputZ = Zoutsig;
	}
	clear();
	move(4, 0);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" PLO   Vdc    Idc      Zin         Zout      Lc      PIM(dBm)       TSSB\n");
	addstr("(dBm)         (mA)   Re    Im    Re    Im   (dB) w1-w2  2w1  2w1-w2   K\n");
	attrset(COLOR_PAIR(COLOR_WHITE));
	sprintf(s, "%5.2f%7.4f%7.3f%6.1f%6.1f%6.1f%6.1f%5.2f%6.1f%6.1f%7.1f%6.1f\n", Plo_dBm, Vdc, 1000.0 * Idc, inputZ.x, inputZ.y, outputZ.x, outputZ.y, loss, PIM2a_dBm, PIM2b_dBm, PIM3_dBm, TSSB);
	addstr(s);
	refresh();
	sleep(1);
	if ((chxx = getch()) != ERR) {
	  addstr("\n");
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Press any key to continue\n");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  refresh();
	  nodelay(stdscr, FALSE);
	  chxx = getch();
	  nodelay(stdscr, TRUE);
	}
	fprintf(sweepfile, "%6.2f%8.4f%7.3f%7.1f%7.1f%7.1f%7.1f%6.2f%6.1f%6.1f%7.1f%7.1f\n", Plo_dBm, Vdc, 1000.0 * Idc, inputZ.x, inputZ.y, outputZ.x, outputZ.y, loss, PIM2a_dBm, PIM2b_dBm, PIM3_dBm, TSSB);
      }
      if (sweepfile != NULL)
	fclose(sweepfile);
      sweepfile = NULL;
      valid_ldata = true;
      valid_sdata = true;
      arpeggio();
    } else if (option == 'V') {
      echo();
      nodelay(stdscr, FALSE);
      clear();
      move(1, 0);
      LO_display = false;
      wants_IM = false;
      wants_conj_data = false;
      addstr("\n");
      refresh();
      do {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" Bias voltage sweep: enter start, stop, step size: ");
	attrset(COLOR_PAIR(COLOR_WHITE));
	ioresult = scanw("%lg%lg%lg", &vstart, &vstop, &vstep);
      } while (ioresult != 3);
      vstep = fabs(vstep);
      if (vstart > vstop)
	vstep = -vstep;
      if (vstep == 0)
	vstep = 1.0e6;
      addstr("\n");
      refresh();
      do {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" Enter input and output frequency indices; n1, n2: ");
	attrset(COLOR_PAIR(COLOR_WHITE));
	ioresult = scanw("%hd%hd", &inpfreq, &outfreq);
      } while (ioresult != 2);
      addstr("\n");
      refresh();
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Filename to save output data: ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      getnstr(swpfilename, 41);
      TEMP1 = strchr(swpfilename, '\n');
      if (TEMP1 != NULL)
	*TEMP1 = 0;
      if (*swpfilename == '\0')
	strcpy(swpfilename, "VBSWEEP");
      if (strchr(swpfilename, '.') == NULL) {
	/* Pascal version uses ".TXT" */
	strcat(swpfilename, ".txt");
	/* possible string truncation in assignment */
      }
      addstr("\n");
      refresh();
      sprintf(s, " Data saved in %s\n", swpfilename);
      addstr(s);
      addstr("\n\n");
      cmplx = zemblo[1];
      Plo_dBm = 10.0 * log(vlosource * vlosource / (8.0 * cmplx.x)) / log(10.0) + 30.0;
      addstr(" Current initial values:\n");
      sprintf(s, "                     Plo, dBm: %7.3f\n", Plo_dBm);
      addstr(s);
      sprintf(s, "   Bias source voltage, Vbias: %7.4f\n\n", vbias);
      addstr(s);
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Want to change anything? ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("(y/n, <cr> to abort): ");
      refresh();
      yes_or_no = getch();
      CAP(&yes_or_no);
      addstr("\n");
      refresh();
      if (yes_or_no == '\015')
	goto _LCONTINUE;
      if (yes_or_no == 'Y') {
	do {
	  addstr("\n");
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Enter new Plo, Vbias: ");
	  refresh();
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  ioresult = scanw("%lg%lg", &Plo_dBm, &vbias);
	} while (ioresult != 2);
      }
      addstr("\n");
      refresh();
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Want IM calculations? ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      addstr("y/n: ");
      refresh();
      yes_or_no = getch();
      CAP(&yes_or_no);
      if (yes_or_no == 'Y') {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr("   Enter input power, dBm: ");
	attrset(COLOR_PAIR(COLOR_WHITE));
	getyx(stdscr, Ypos, Xpos);
	do {
	  is_OK = true;
	  ioresult = scanw("%lg", &Pin_dBm);
	  if (ioresult != 1) {
	    is_OK = false;
	    move(Ypos, Xpos);
	    clrtoeol();
	  }
	} while (!is_OK);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	input_power = 0.001 * X_tothe_Y(10.0, Pin_dBm / 10.0);
	wants_IM = true;
      } else {
	addstr("\n");
      }
      addstr("\n");
      refresh();
      attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
      addstr(" Actual or Conjugate-match loss and impedances (a/c): ");
      attrset(COLOR_PAIR(COLOR_WHITE));
      A_or_C = getch();
      CAP(&A_or_C);
      addstr("\n");
      refresh();
      if (A_or_C == 'C')
	wants_conj_data = true;
      strcpy(sweepfile_NAME, swpfilename);
      if (sweepfile != NULL)
	sweepfile = freopen(sweepfile_NAME, "w", sweepfile);
      else
	sweepfile = fopen(sweepfile_NAME, "w");
      if (sweepfile == NULL)
	error_exit(FileNotFound, sweepfile_NAME);
      fprintf(sweepfile, "\n BIAS VOLTAGE SWEEP          DATA FILE IS %s\n\n", datafilename);
      fprintf(sweepfile, " INPUT FREQ:%8.4f GHz   OUTPUT FREQ:%8.4f GHz   LO FREQ:%8.4f GHz\n\n", 1.0e-9 * fabs(inpfreq * LOfreq + IFfreq), 1.0e-9 * fabs(outfreq * LOfreq + IFfreq), 1.0e-9 * LOfreq);
      if (wants_IM)
	fprintf(sweepfile, " RF INPUT POWER= %6.2f dBm\n\n", Pin_dBm);
      if (wants_conj_data)
	fprintf(sweepfile, " Tabulated results are CONJUGATE MATCH values\n");
      else
	fprintf(sweepfile, " Tabulated results are ACTUAL values\n");
      fprintf(sweepfile, "\n  PLO    Vdc    Idc        Zin          Zout        Lc      PIM(dBm)        TSSB\n");
      fprintf(sweepfile, " (dBm)          (mA)    Re     Im     Re     Im    (dB) w1-w2  2w1  2w1-w2 deg. K\n");
      putc(' ', sweepfile);
      for (j = 2; j <= 81; j++)
	putc('=', sweepfile);
      fprintf(sweepfile, "\n\n");
      pavail = 0.001 * X_tothe_Y(10.0, Plo_dBm / 10.0);
      vlosource = sqrt(8.0 * cmplx.x * pavail);
      steps = labs((long) floor((vstop - vstart) / vstep + 0.5));
      FORLIM = steps;
      for (stepnum = 0; stepnum <= FORLIM; stepnum++) {
	Vdc_desired = vstart + vstep * stepnum;
	if (stepnum != 0)
	  vbias += Vdc_desired - Vdc;
	vlosource = sqrt(8.0 * cmplx.x * pavail);
	do_lgsg(vlosource, &vbias);
	move(7, 36);
	attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	addstr("PERFORMING SMALL-SIG ANALYSIS\n");
	refresh();
	attrset(COLOR_PAIR(COLOR_WHITE));
	Small_sig(c, ij, LOfreq, IFfreq, Yconv, Zconv, JYconv, JZconv);
	SSPrams(inpfreq, outfreq, Yconv, &Zinsig, &Zoutsig, &Zinmatch, &Zoutmatch, &conv_loss, &rev_loss, &MAGain);
	NOISE(JYconv, zembsig, Zinmatch, Zoutmatch, Zoutsig, inpfreq, outfreq, conv_loss, MAGain, ij, wants_conj_data, &TSSB, &Tout);
	if (wants_IM) {
	  move(8, 36);
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("   PERFORMING IM ANALYSIS\n");
	  refresh();
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  IM(inpfreq, outfreq, Zconv, Yconv, JZconv, JYconv, LOfreq, IFfreq, IFfreq2, input_power, vj, ij, c, zembsig, Zinmatch, Zoutmatch, wants_conj_data, &IM2a_power, &IM2b_power, &IM3_power);
	  PIM2a_dBm = 10.0 * log(IM2a_power) / log(10.0) + 30.0;
	  PIM2b_dBm = 10.0 * log(IM2b_power) / log(10.0) + 30.0;
	  PIM3_dBm = 10.0 * log(IM3_power) / log(10.0) + 30.0;
	} else {
	  PIM2a_dBm = 0.0;
	  PIM2b_dBm = 0.0;
	  PIM3_dBm = 0.0;
	}
	if (wants_conj_data) {
	  loss = MAGain;
	  inputZ = Zinmatch;
	  outputZ = Zoutmatch;
	} else {
	  loss = conv_loss;
	  inputZ = Zinsig;
	  outputZ = Zoutsig;
	}
	clear();
	move(5, 0);
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	addstr(" PLO   Vdc    Idc      Zin         Zout      Lc      PIM(dBm)       TSSB\n");
	addstr("(dBm)         (mA)   Re    Im    Re    Im   (dB) w1-w2  2w1  2w1-w2   K\n");
	attrset(COLOR_PAIR(COLOR_WHITE));
	sprintf(s, "%5.2f%7.4f%7.3f%6.1f%6.1f%6.1f%6.1f%5.2f%6.1f%6.1f%7.1f%6.1f\n", Plo_dBm, Vdc, 1000.0 * Idc, inputZ.x, inputZ.y, outputZ.x, outputZ.y, loss, PIM2a_dBm, PIM2b_dBm, PIM3_dBm, TSSB);
	addstr(s);
	refresh();
	sleep(1);
	if ((chxx = getch()) != ERR) {
	  addstr("\n");
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr(" Press any key to continue\n");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  refresh();
	  nodelay(stdscr, FALSE);
	  chxx = getch();
	  nodelay(stdscr, TRUE);
	}
	fprintf(sweepfile, "%6.2f%8.4f%7.3f%7.1f%7.1f%7.1f%7.1f%6.2f%6.1f%6.1f%7.1f%7.1f\n", Plo_dBm, Vdc, 1000.0 * Idc, inputZ.x, inputZ.y, outputZ.x, outputZ.y, loss, PIM2a_dBm, PIM2b_dBm, PIM3_dBm, TSSB);
      }
      if (sweepfile != NULL)
	fclose(sweepfile);
      sweepfile = NULL;
      valid_ldata = true;
      valid_sdata = true;
      arpeggio();
    } else if (option == 'G') {
      for (j = 0; j <= 31; j++)
	ic[j] = id[j] - ij[j];
      echo();
      nodelay(stdscr, FALSE);
      if (valid_ldata) {
	do {
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  clear();
	  FRAME();
	  move(3, 0);
	  addstr("        Graph types:\n");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr("\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    vj");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": diode junction voltage\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    vd");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": diode terminal voltage\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    ij");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": resistive junction current\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    ic");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": capacitive junction current\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    id");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": total junction current\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    cj");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": junction capacitance\n");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  addstr("                    qq");
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  addstr(": quit and return to main menu\n\n");
	  attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	  addstr("        Enter graph type: ");
	  attrset(COLOR_PAIR(COLOR_YELLOW) | A_BOLD);
	  refresh();
	  getnstr(response, 41);
	  TEMP1 = strchr(response, '\n');
	  if (TEMP1 != NULL)
	    *TEMP1 = 0;
	  attrset(COLOR_PAIR(COLOR_WHITE));
	  RemoveSpaces(response);
	  sprintf(response, "%.2s", strcpy(STR2, response));
	  /* possible string truncation in assignment */
	  CAP(response);
	  CAP(&response[1]);
	  if (!strcmp(response, "VJ"))
	    drawgraph(juncv, vj);
	  if (!strcmp(response, "VD"))
	    drawgraph(diodev, vd);
	  if (!strcmp(response, "IJ"))
	    drawgraph(junccur, ij);
	  if (!strcmp(response, "ID"))
	    drawgraph(diodecur, id);
	  if (!strcmp(response, "IC"))
	    drawgraph(capcur, ic);
	  if (!strcmp(response, "CJ"))
	    drawgraph(junccap, c);
	} while (strcmp(response, "QQ"));
      } else {
	WINDOW *tmp_win;
	tmp_win = subwin(stdscr, 2, 58, 11, 18);
	wbkgd(tmp_win, COLOR_PAIR(COLOR_RED));
	wclear(tmp_win);
	wattrset(tmp_win, COLOR_PAIR(COLOR_RED));
	wmove(tmp_win, 0, 0);
	waddstr(tmp_win, "No valid large-signal data; run LGSG or read from a file\n");
	waddstr(tmp_win, "Strike any key to continue");
	touchwin(stdscr);
	wrefresh(tmp_win);
	NoError = false;
	chxx = getch();
	delwin(tmp_win);
      }
      noecho();
      nodelay(stdscr, TRUE);
    }				/*menu option selection */
    if (option != 'G' && option != 'Q') {
      if (NoError) {
	attrset(COLOR_PAIR(COLOR_WHITE) | A_BOLD);
	move(22, 0);
	clrtoeol();
	addstr(" Strike any key to continue");
	refresh();
	attrset(COLOR_PAIR(COLOR_WHITE));
      }
      nodelay(stdscr, FALSE);
      ch3 = getch();
      nodelay(stdscr, TRUE);
    }
  _LCONTINUE:
  } while (option != 'Q');
  if (opened_soutfile) {
    if (soutfile != NULL)
      fclose(soutfile);
    soutfile = NULL;
  }
_LFINIS:
  
  endwin();

  if (infile != NULL)
    fclose(infile);
  if (soutfile != NULL)
    fclose(soutfile);
  if (loutfile != NULL)
    fclose(loutfile);
  if (vjijfile != NULL)
    fclose(vjijfile);
  if (sweepfile != NULL)
    fclose(sweepfile);
  exit(0);
}
