/* dmxsmall.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: DMXSMALL.PAS */

#include "cmplxops.h"
#include "diodemx.h"
#include "cinvert.h"
#include "dmxsmall.h"
#include "fft.h"


extern double delta;
extern Zsigtype zembsig;
extern complex one, zero;
extern double rs;

/**************** SMALL-SIGNAL ANALYSIS PROCEDURES ****************/


void Small_sig(waveform c, waveform ij, double LOfreq, double IFfreq, complex(*Yconv)[9], complex(*Zconv)[9], complex(*JYconv)[9], complex(*JZconv)[9])
{
  short j, k;
  complex cmplx, cmplx1;
  waveform gj;
  freqvector cfreal, cfimag, gjfreal, gjfimag;
  complex omega[9];
  complex cfreq[17], gjfreq[17];

  /* Form g(t) waveform, gj */
  for (j = 0; j <= 31; j++)
    gj[j] = delta * ij[j];

  /* FFT c(t) and gj(t) to get frequency components */
  FFT(c, cfreal, cfimag, 32, 5, 1);
  FFT(gj, gjfreal, gjfimag, 32, 5, 1);

  /* Form complex Fourier coefs of freq-domain jcn conductance &
     capacitance. Store as vectors [-8..8] and
     throw away higher harmonics */
  cmplx.x = 0.5 * cfreal[0];
  cmplx.y = 0.0;
  cfreq[8] = cmplx;
  cmplx.x = 0.5 * gjfreal[0];
  cmplx.y = 0.0;
  gjfreq[8] = cmplx;
  for (j = 1; j <= 8; j++) {
    cmplx.x = 0.5 * cfreal[j];
    cmplx.y = 0.5 * cfimag[j];
    cfreq[j + 8] = cmplx;
    cmplx.y = -cmplx.y;
    cfreq[8 - j] = cmplx;

    cmplx.x = 0.5 * gjfreal[j];
    cmplx.y = 0.5 * gjfimag[j];
    gjfreq[j + 8] = cmplx;
    cmplx.y = -cmplx.y;
    gjfreq[8 - j] = cmplx;
  }

  /* Form omega, the vector of mixing frequencies */
  /* omega is complex to aid in multiplications, later */
  /* Note that LSB values of omega are negative */
  for (j = -4; j <= 4; j++) {
    cmplx.x = 0.0;
    cmplx.y = 2.0 * M_PI * (IFfreq + j * LOfreq);
    omega[j + 4] = cmplx;
  }

  /* Form Zconv, the diode conversion matrix */
  /* Note that it is [-4..4,-4..4] to match mixing freq terms */
  for (j = 0; j <= 8; j++) {	/* row no. */
    for (k = 0; k <= 8; k++) {	/* col. no. */
      cmplx = CMULT(cfreq[j - k + 8], omega[j]);
      Yconv[j][k] = CADD(cmplx, gjfreq[j - k + 8]);
    }
  }

  Cinvert(Yconv, Zconv);
  memcpy(JYconv, Yconv, sizeof(cmplx_mtx));
  memcpy(JZconv, Zconv, sizeof(cmplx_mtx));

  /* Augment with Rs+Ze */
  cmplx1.x = rs;
  cmplx1.y = 0.0;
  for (j = 0; j <= 8; j++) {
    cmplx = CADD(zembsig[j], cmplx1);
    Zconv[j][j] = CADD(cmplx, Zconv[j][j]);
  }

  /* Finally invert to get Yconv */
  Cinvert(Zconv, Yconv);
}


void SSPrams(short inpfreq, short outfreq, complex(*Yconv)[9], complex * Zinsig, complex * Zoutsig, complex * Zinmatch, complex * Zoutmatch, double *conv_loss, double *rev_loss, double *MAGain)
{

  /* Calculates gain and input/output impedances of the mixer.
     All impedances output are conventional phasor form.
     conv_loss, Gain are in dB; + for loss */
  complex cmplx, cmplx1, cmplx2, cmplx3, cmplx4, det, yf, yr, yi, yo;
  cmplx_mtx Z2X2;
  double arg, mag, phase, Rload, Rsource, TEMP1;


  /* Calc. conversion loss */
  /* Get magnitude of Yconv component */
  COM_TO_POLAR(Yconv[outfreq + 4][inpfreq + 4], &mag, &phase);

  /* Get real parts of source & load Z */
  cmplx = zembsig[inpfreq + 4];
  Rsource = cmplx.x;
  cmplx = zembsig[outfreq + 4];
  Rload = cmplx.x;

  /* Calc. conv_loss in dB */
  *conv_loss = -10.0 * log(4.0 * mag * mag * Rsource * Rload) / log(10.0);

  /* Calc. reverse conv. loss the same way */
  COM_TO_POLAR(Yconv[inpfreq + 4][outfreq + 4], &mag, &phase);
  *rev_loss = -10.0 * log(4.0 * mag * mag * Rsource * Rload) / log(10.0);

  /* Calc. input impedance */
  cmplx = CDIV(one, Yconv[inpfreq + 4][inpfreq + 4]);
  *Zinsig = CSUB(cmplx, zembsig[inpfreq + 4]);
  if (inpfreq < 0)
    (*Zinsig).y = -(*Zinsig).y;

  /* Calc. output impedance */
  cmplx = CDIV(one, Yconv[outfreq + 4][outfreq + 4]);
  *Zoutsig = CSUB(cmplx, zembsig[outfreq + 4]);
  if (outfreq < 0)
    (*Zoutsig).y = -(*Zoutsig).y;

  /* Calc simul. conj match impedances and MAG */
  if (inpfreq != outfreq) {
    /* 1. generate 2X2 Y-matrix without embedding impedances at inp/out */
    /*    and put in standard two-port form */
    /* Start by inverting 2X2 Yconv to get Z2X2, the 2X2 Z-matrix */
    cmplx = CMULT(Yconv[inpfreq + 4][inpfreq + 4], Yconv[outfreq + 4][outfreq + 4]);
    cmplx1 = CMULT(Yconv[inpfreq + 4][outfreq + 4], Yconv[outfreq + 4][inpfreq + 4]);
    det = CSUB(cmplx, cmplx1);
    /*det is the determinant */
    Z2X2[outfreq + 4][outfreq + 4] = CDIV(Yconv[inpfreq + 4][inpfreq + 4], det);
    Z2X2[inpfreq + 4][outfreq + 4] = CDIV(Yconv[inpfreq + 4][outfreq + 4], det);
    Z2X2[inpfreq + 4][outfreq + 4] = CSUB(zero, Z2X2[inpfreq + 4][outfreq + 4]);
    Z2X2[outfreq + 4][inpfreq + 4] = CDIV(Yconv[outfreq + 4][inpfreq + 4], det);
    Z2X2[outfreq + 4][inpfreq + 4] = CSUB(zero, Z2X2[outfreq + 4][inpfreq + 4]);
    Z2X2[inpfreq + 4][inpfreq + 4] = CDIV(Yconv[outfreq + 4][outfreq + 4], det);

    /* Subtract Ze at input and output */
    Z2X2[inpfreq + 4][inpfreq + 4] = CSUB(Z2X2[inpfreq + 4][inpfreq + 4], zembsig[inpfreq + 4]);
    Z2X2[outfreq + 4][outfreq + 4] = CSUB(Z2X2[outfreq + 4][outfreq + 4], zembsig[outfreq + 4]);

    /* Z2X2 is now the two-port Z matrix of the diode, augmented
       at all freqs except input and output */
    /* Invert to get a Y-matrix */
    cmplx = CMULT(Z2X2[inpfreq + 4][inpfreq + 4], Z2X2[outfreq + 4][outfreq + 4]);
    cmplx1 = CMULT(Z2X2[outfreq + 4][inpfreq + 4], Z2X2[inpfreq + 4][outfreq + 4]);
    det = CSUB(cmplx, cmplx1);
    yo = CDIV(Z2X2[inpfreq + 4][inpfreq + 4], det);
    yr = CDIV(Z2X2[inpfreq + 4][outfreq + 4], det);
    yr = CSUB(zero, yr);
    yf = CDIV(Z2X2[outfreq + 4][inpfreq + 4], det);
    yf = CSUB(zero, yf);
    yi = CDIV(Z2X2[outfreq + 4][outfreq + 4], det);

    /*yi,yo,yf,yr are now the two-port Y-params in Carson's notation */
    /* 2. Calculate simul. conj. match impedances */
    cmplx = CMULT(yr, yf);
    COM_TO_POLAR(cmplx, &mag, &phase);
    TEMP1 = 2.0 * yi.x * yo.x - cmplx.x;
/* p2c: dmxsmall.pas: Note: Eliminated unused assignment statement [338] */
    arg = TEMP1 * TEMP1 - mag * mag;
    if (arg < 0) {		/* i.e. conditionally stable */
/* p2c: dmxsmall.pas: Note: Eliminated unused assignment statement [338] */
      *Zinmatch = zero;
      *Zoutmatch = zero;
      *MAGain = 0.00;		/*dB */
      return;
    }
    TEMP1 = 2.0 * yi.x * yo.x - cmplx.x;
    (*Zinmatch).x = 0.5 * sqrt(TEMP1 * TEMP1 - mag * mag) / yo.x;
    (*Zinmatch).y = yi.y - 0.5 * cmplx.y / yo.x;
    (*Zinmatch) = CDIV(one, *Zinmatch);
    TEMP1 = 2.0 * yi.x * yo.x - cmplx.x;
    (*Zoutmatch).x = 0.5 * sqrt(TEMP1 * TEMP1 - mag * mag) / yi.x;
    (*Zoutmatch).y = yo.y - 0.5 * cmplx.y / yi.x;
    (*Zoutmatch) = CDIV(one, *Zoutmatch);

    /* 3. Calculate MAG */
    /*cmplx is still yr*yf */
    cmplx1 = *Zinmatch;
    cmplx1.y = -cmplx1.y;	/*cmplx1 is now the source imp. */
    cmplx2 = *Zoutmatch;
    cmplx2.y = -cmplx2.y;	/*cmplx2 is now the load imp. */
    cmplx1 = CDIV(one, cmplx1);
    cmplx2 = CDIV(one, cmplx2);
    cmplx3 = CADD(yi, cmplx1);
    cmplx4 = CADD(yo, cmplx2);
    cmplx3 = CMULT(cmplx3, cmplx4);
    cmplx = CSUB(cmplx3, cmplx);
    /*cmplx is now the sqrt of the denom. of Gt */
    COM_TO_POLAR(cmplx, &mag, &phase);
    *MAGain = 4.0 * cmplx1.x * cmplx2.x * (yf.x * yf.x + yf.y * yf.y) / (mag * mag);
    *MAGain = -10.0 * log10(*MAGain);

    /* Conjugate the inp/out impedance for LSB freqs */
    if (inpfreq < 0)
      (*Zinmatch).y = -(*Zinmatch).y;
    if (outfreq < 0)
      (*Zoutmatch).y = -(*Zoutmatch).y;
    return;
  }
/* p2c: dmxsmall.pas: Note: Eliminated unused assignment statement [338] */
  *Zinmatch = zero;
  *Zoutmatch = zero;
  *MAGain = 0.0;
  /*if outfreq=inpfreq */
}
