/* fft.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: FFT.PAS */

#include <math.h>
#include "fft.h"

short j_to_the_k(short j, short k)
{
  short x = 1;
  short i;

  for (i = 1; i <= k; i++)
    x *= j;
  return x;
}

short ibitr(short j, short nu)
{
  short j1 = j;
  short j2, i;
  short x = 0;

  for (i = 1; i <= nu; i++) {
    j2 = j1 / 2;
    x = x * 2 + j1 - j2 * 2;
    j1 = j2;
  }
  return x;
}


void FFT(double *tvct, double *fvctr, double *fvcti, short n, short nu, short kflag)
{

  /*  Sin/Cos FFT subroutine.  Ref: Ch. 10, "The Fast Fourier Transform"
     by E. Oran Brigham (Prentice-Hall, 1974).

     fvctr is the cosine (a) coefficient; fvcti is the sin (b) coefficient.
     Form of the series is

     f(t)=a0/2 + a1*cos(wt) - b1*sin(wt) + a2*cos(2wt) - b2*sin(2wt) + ...

     and fvctr(0)=DC component, fvctr(n)=nth harmonic, etc.  Components must
     be entered with these sign conventions and twice the DC value for a0, as
     implied above.  kflag=1 implies time to freq conversion; kflag<>1 implies
     freq to time conversion.  waveform is array[0..31] of double;
     freqvector is array[0..16] of double.

     n is the number of points in the time waveform (32, max) and must be a
     power of 2.  nu=log2(n). */
  short m, n2, n2p1, nu1, jj, jk;
  short k = 0;
  short p, k1, k1n2, l, i, loop;
  double treal, timag, c, s, coef, arg;
  double xreal[32], ximag[32];

  n2 = n / 2;
  n2p1 = n2 + 1;
  if (kflag == 1) {
    for (jj = 0; jj <= n - 1; jj++) {
      xreal[jj] = tvct[jj];
      ximag[jj] = 0.0;
    }
  } else {
    for (jj = 1; jj <= n2; jj++) {
      jk = n - jj + 1;
      xreal[jj - 1] = 0.5 * fvctr[jj - 1];
      xreal[jk - 1] = 0.5 * fvctr[jj];
      ximag[jj - 1] = -0.5 * fvcti[jj - 1];
      ximag[jk - 1] = 0.5 * fvcti[jj];
    }
    ximag[0] = 0.0;
    ximag[n2p1 - 1] = 0.0;
  }
  nu1 = nu - 1;
  for (l = 1; l <= nu; l++) {
    do {
      for (i = 1; i <= n2; i++) {
	m = j_to_the_k(2, nu1);
	p = ibitr(k / m, nu);
	arg = p * 6.283185307 / n;
	c = cos(arg);
	s = sin(arg);
	k1 = k + 1;
	k1n2 = k1 + n2;
	treal = xreal[k1n2 - 1] * c + ximag[k1n2 - 1] * s;
	timag = ximag[k1n2 - 1] * c - xreal[k1n2 - 1] * s;
	xreal[k1n2 - 1] = xreal[k1 - 1] - treal;
	ximag[k1n2 - 1] = ximag[k1 - 1] - timag;
	xreal[k1 - 1] += treal;
	ximag[k1 - 1] += timag;
	k++;
      }
      k += n2;
    } while (k < n);
    k = 0;
    nu1--;
    n2 /= 2;
  }
  for (k = 0; k <= n - 1; k++) {
    i = ibitr(k, nu) + 1;
    if (i > k + 1) {
      treal = xreal[k];
      timag = ximag[k];
      xreal[k] = xreal[i - 1];
      ximag[k] = ximag[i - 1];
      xreal[i - 1] = treal;
      ximag[i - 1] = timag;
    }
  }
  if (kflag != 1) {
    for (loop = 0; loop <= n - 1; loop++)
      tvct[loop] = xreal[loop];
    return;
  }
  coef = 2.0 / n;
  for (loop = 0; loop <= n2p1 - 1; loop++) {
    fvctr[loop] = xreal[loop] * coef;
    fvcti[loop] = ximag[loop] * coef;
  }
  fvcti[n2p1 - 1] = -fvcti[n2p1 - 1];
}
