/* im.c
 * 
 * Copyright (C) 2002 Claudio Girardi  <in3otd@qsl.net>
 * 
 * This file was derived from the original Pascal code, written by
 * Stephen A. Maas. It was translated into ANSI C with the help of p2c,
 * version 1.21alpha, and further manually refined by Claudio Girardi.
 * 
 * Bugs were certainly introduced during the translation process, so
 * please don't bother Sthephen with bug reports but send them to
 * Claudio, thanks.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA, 02111-1307, USA
 */


/* Original file: IM.PAS */


#include "cmplxops.h"
#include "diodemx.h"
#include "cinvert.h"
#include "fft.h"
#include "im.h"

extern double delta;
extern double phi, rs;

void IM(short inpfreq, short outfreq, complex(*Zconv_)[9], complex(*Yconv_)[9], complex(*JZconv_)[9], complex(*JYconv_)[9], double LOfreq, double IFfreq, double IFfreq2, double input_power, double *vj_, double *ij_, double *c_, complex * zembsig_, complex Zinmatch_, complex Zoutmatch_, boolean conjmatch, double *IM2a_power, double *IM2b_power, double *IM3_power)
{
  cmplx_mtx Zconv, Yconv, JZconv, JYconv;
  waveform vj, ij, c;
  Zsigtype zembsig;
  complex Zinmatch, Zoutmatch;

  short a, b, l, m, n, p;
  waveform f;
  double delta_coef, phase, IdIMmag;
  complex Vsig, cmplx, cmplx1, two, one, zero, quarter, half, inpzemb, outzemb, delta_cmplx, IdIM3;
  freqvector Fcos, Fsin;
  complex C2freq[25], C3freq[25];
  complex IM2aterm[9][17], IM2bterm[9][17];
  complex IM23term[9][25], IM3term[9][25];
  complex ILOfreq[25], ILO3freq[25];
  cmplx_mtx GZmtx;
  complex Vtran[9], V1freqminus[9], V1freqplus[9], zsigRs[9], IM2aSC[9], IM2bSC[9], Vj2a[9], Vj2b[9], IdIM2a[9], IdIM2b[9], IM23SC[9], IM3SC[9];
  boolean stable;
  double TEMP1;

  /*Any variable with the term 2a in it refers to the w1-w2 frequency
     components; 2b refers to the 2*w1 components */


  memcpy(Zconv, Zconv_, sizeof(cmplx_mtx));
  memcpy(Yconv, Yconv_, sizeof(cmplx_mtx));
  memcpy(JZconv, JZconv_, sizeof(cmplx_mtx));
  memcpy(JYconv, JYconv_, sizeof(cmplx_mtx));
  memcpy(vj, vj_, sizeof(waveform));
  memcpy(ij, ij_, sizeof(waveform));
  memcpy(c, c_, sizeof(waveform));
  memcpy(zembsig, zembsig_, sizeof(Zsigtype));
  Zinmatch = Zinmatch_;
  Zoutmatch = Zoutmatch_;

  /*define complex constants */
  two.x = 2.0;
  two.y = 0.0;
  one.x = 1.0;
  one.y = 0.0;
  zero.x = 0.0;
  zero.y = 0.0;
  half.x = 0.5;
  half.y = 0.0;
  quarter.x = 0.25;
  quarter.y = 0.0;

  /*Check stability and bail out if unstable & conj. match data is desired */
  stable = !(conjmatch && Zinmatch.y == 0.0 && Zoutmatch.y == 0.0);
  if (!stable) {
    *IM2a_power = 0.001;
    *IM2b_power = 0.001;
    *IM3_power = 0.001;
    goto _L625;			/*end of procedure */
  }
  if (conjmatch) {		/*Replace zembsig with Zin/out_match at RF & IF */
    Zconv[inpfreq + 4][inpfreq + 4] = CSUB(Zconv[inpfreq + 4][inpfreq + 4], zembsig[inpfreq + 4]);
    Zconv[outfreq + 4][outfreq + 4] = CSUB(Zconv[outfreq + 4][outfreq + 4], zembsig[outfreq + 4]);
    if (inpfreq < 0)
      Zinmatch = CONJ(Zinmatch);
    if (outfreq < 0)
      Zoutmatch = CONJ(Zoutmatch);
    zembsig[inpfreq + 4] = CONJ(Zinmatch);
    zembsig[outfreq + 4] = CONJ(Zoutmatch);
    Zconv[inpfreq + 4][inpfreq + 4] = CADD(Zconv[inpfreq + 4][inpfreq + 4], zembsig[inpfreq + 4]);
    Zconv[outfreq + 4][outfreq + 4] = CADD(Zconv[outfreq + 4][outfreq + 4], zembsig[outfreq + 4]);
    Cinvert(Zconv, Yconv);
  }
  inpzemb = zembsig[inpfreq + 4];
  outzemb = zembsig[outfreq + 4];
  delta_coef = 0.5 * delta * delta;
  delta_cmplx.x = delta;
  delta_cmplx.y = 0.0;

  /*Find junction current Fourier coefs and mult by d**2/2,=ILOfreq */
  FFT(ij, Fcos, Fsin, 32, 5, 1);
  cmplx.x = 0.5 * delta_coef * Fcos[0];
  cmplx.y = 0.0;
  ILOfreq[12] = cmplx;
  for (a = 1; a <= 12; a++) {
    cmplx.x = 0.5 * delta_coef * Fcos[a];
    cmplx.y = 0.5 * delta_coef * Fsin[a];
    ILOfreq[12 + a] = cmplx;
    cmplx.y = -cmplx.y;
    ILOfreq[12 - a] = cmplx;
  }

  /*Calculate IM2 capacitive coefficient waveform */
  for (a = 0; a <= 31; a++)
    f[a] = 0.25 * c[a] / (phi - vj[a]);

  /*Find C2 Fourier coefficients */
  FFT(f, Fcos, Fsin, 32, 5, 1);
  cmplx.x = 0.5 * Fcos[0];
  cmplx.y = 0.0;
  C2freq[12] = cmplx;
  for (a = 1; a <= 12; a++) {
    cmplx.x = 0.5 * Fcos[a];
    cmplx.y = 0.5 * Fsin[a];
    C2freq[a + 12] = cmplx;
    cmplx.y = -cmplx.y;
    C2freq[12 - a] = cmplx;
  }

  /*Find IM2_term, the IM generation coefficient */
  cmplx.x = 0.0;
  cmplx1.x = 0.0;
  for (a = 0; a <= 8; a++) {
    cmplx.y = 2.0 * M_PI * ((a - 4) * LOfreq + fabs(IFfreq) - fabs(IFfreq2));
    cmplx1.y = 2.0 * M_PI * ((a - 4) * LOfreq + 2.0 * fabs(IFfreq));
    for (l = -8; l <= 8; l++) {
      IM2aterm[a][l + 8] = CMULT(C2freq[l + 12], cmplx);
      IM2aterm[a][l + 8] = CADD(ILOfreq[l + 12], IM2aterm[a][l + 8]);
      IM2bterm[a][l + 8] = CMULT(C2freq[l + 12], cmplx1);
      IM2bterm[a][l + 8] = CADD(ILOfreq[l + 12], IM2bterm[a][l + 8]);
    }
  }

  /*Find first order voltages */

  /*1. Find signal voltage magnitude */
  Vsig.x = sqrt(8.0 * input_power * inpzemb.x);
  Vsig.y = 0.0;

  /*2. Find transfer function between Vsig and V1freq, Vtran, column
     of a conversion matrix */
  for (a = 0; a <= 8; a++) {
    cmplx = zero;
    for (b = 0; b <= 8; b++) {
      cmplx1 = CMULT(JZconv[a][b], Yconv[b][inpfreq + 4]);
      cmplx = CADD(cmplx, cmplx1);
    }
    Vtran[a] = cmplx;
  }

  /*3. Multiply Vtran by Vsig to obtain V1freq (first order components) */
  for (a = -4; a <= 4; a++) {
    cmplx = CMULT(Vtran[a + 4], Vsig);
    V1freqplus[4 + a] = cmplx;	/*w1 or w2 components */
    cmplx.y = -cmplx.y;
    V1freqminus[4 - a] = cmplx;
	/*-w1 or -w2 components*/
  }

  /*Add Rs to zembsig to get zsigRs */
  cmplx.x = rs;
  cmplx.y = 0.0;
  for (a = 0; a <= 8; a++)
    zsigRs[a] = CADD(zembsig[a], cmplx);

  /*Form the GZ matrix; GZ=(1+JYconv*ZsigRs)**-1 */
  for (a = 0; a <= 8; a++) {
    for (b = 0; b <= 8; b++) {
      GZmtx[a][b] = CMULT(JYconv[a][b], zsigRs[b]);
      if (a == b)
	GZmtx[a][b] = CADD(GZmtx[a][b], one);
    }
  }
  Cinvert(GZmtx, GZmtx);

  /*Calc. IM2_SC vectors */
  for (a = 0; a <= 8; a++) {
    IM2aSC[a] = zero;
    IM2bSC[a] = zero;
  }
  for (l = 0; l <= 16; l++) {
    for (m = -4; m <= 4; m++) {
      for (p = m; p <= 4; p++) {
	a = m + p + l - 8;
	if (abs(a) <= 4) {
	  cmplx = CMULT(V1freqminus[p + 4], V1freqplus[m + 4]);
	  if (m != p) {
	    cmplx1 = CMULT(V1freqminus[m + 4], V1freqplus[p + 4]);
	    cmplx = CADD(cmplx, cmplx1);
	  }
	  cmplx = CMULT(IM2aterm[a + 4][l], cmplx);
	  IM2aSC[a + 4] = CADD(IM2aSC[a + 4], cmplx);
	  cmplx = CMULT(V1freqplus[p + 4], V1freqplus[m + 4]);
	  if (m != p)
	    cmplx = CMULT(two, cmplx);
	  cmplx = CMULT(IM2bterm[a + 4][l], cmplx);
	  IM2bSC[a + 4] = CADD(IM2bSC[a + 4], cmplx);
	}
      }				/*p */
    }				/*m */
  }				/*l */
  for (a = 0; a <= 8; a++)
    IM2bSC[a] = CMULT(half, IM2bSC[a]);

  /*Multiply GZmtx by IM2_SC to get the IM output currents, IdIM2_;
     also find jcn voltages Vj2_ */
  for (a = 0; a <= 8; a++) {
    IdIM2a[a] = zero;
    IdIM2b[a] = zero;
    for (b = 0; b <= 8; b++) {
      cmplx = CMULT(GZmtx[a][b], IM2aSC[b]);
      IdIM2a[a] = CADD(cmplx, IdIM2a[a]);
      cmplx = CMULT(GZmtx[a][b], IM2bSC[b]);
      IdIM2b[a] = CADD(cmplx, IdIM2b[a]);
    }
    Vj2a[a] = CMULT(zsigRs[a], IdIM2a[a]);
    Vj2a[a] = CSUB(zero, Vj2a[a]);
    Vj2b[a] = CMULT(zsigRs[a], IdIM2b[a]);
    Vj2b[a] = CSUB(zero, Vj2b[a]);
  }

  /*Finally, calculate IM2 powers */
  COM_TO_POLAR(IdIM2a[outfreq + 4], &IdIMmag, &phase);
  *IM2a_power = 0.5 * IdIMmag * IdIMmag * outzemb.x;	/*w1-w2 term */
  COM_TO_POLAR(IdIM2b[outfreq + 4], &IdIMmag, &phase);
  *IM2b_power = 0.5 * IdIMmag * IdIMmag * outzemb.x;	/*2*w1 term */

  /*IM3 CALCULATION */
  /*Calculate IM3 capacitive coefficient waveform */
  for (a = 0; a <= 31; a++) {
    TEMP1 = phi - vj[a];
    f[a] = 0.125 * c[a] / (TEMP1 * TEMP1);
  }

  /*Find C3 Fourier coefficients */
  FFT(f, Fcos, Fsin, 32, 5, 1);
  cmplx.x = 1.5 * Fcos[0];	/*the 1.5 includes the factor of 3 */
  cmplx.y = 0.0;
  C3freq[12] = cmplx;
  for (a = 1; a <= 12; a++) {
    cmplx.x = 1.5 * Fcos[a];
    cmplx.y = 1.5 * Fsin[a];
    C3freq[a + 12] = cmplx;
    cmplx.y = -cmplx.y;
    C3freq[12 - a] = cmplx;
  }

  /*calc. ILO for cubic summation */
  for (l = 0; l <= 24; l++)
    ILO3freq[l] = CMULT(delta_cmplx, ILOfreq[l]);

  /*calculate IM_3term, the IM3 generating coefs */
  cmplx.x = 0.0;
  for (a = 0; a <= 8; a++) {
    cmplx.y = 2.0 * M_PI * ((a - 4) * LOfreq + 2.0 * fabs(IFfreq) - fabs(IFfreq2));
    for (l = 0; l <= 24; l++) {
      IM23term[a][l] = CMULT(C2freq[l], cmplx);
      IM23term[a][l] = CADD(ILOfreq[l], IM23term[a][l]);
      IM3term[a][l] = CMULT(C3freq[l], cmplx);
      IM3term[a][l] = CADD(ILO3freq[l], IM3term[a][l]);
    }
  }

  /*Calc. IM23SC vector, the SC IM current due to 1st & 2nd order voltages */
  for (a = 0; a <= 8; a++)
    IM23SC[a] = zero;
  for (l = -12; l <= 12; l++) {
    for (m = -4; m <= 4; m++) {
      for (p = 0; p <= 8; p++) {
	a = m + p + l - 4;
	if (abs(a) <= 4) {
	  cmplx = CMULT(Vj2a[p], V1freqplus[m + 4]);
	  cmplx1 = CMULT(Vj2b[p], V1freqminus[m + 4]);
	  cmplx = CADD(cmplx1, cmplx);
	  cmplx = CMULT(IM23term[a + 4][l + 12], cmplx);
	  IM23SC[a + 4] = CADD(cmplx, IM23SC[a + 4]);
	}
      }				/*p */
    }				/*m */
  }				/*l */

  /*Calc V3sigma vector */
  for (a = 0; a <= 8; a++)
    IM3SC[a] = zero;
  for (l = -12; l <= 12; l++) {
    for (m = -4; m <= 4; m++) {
      for (n = m; n <= 4; n++) {
	for (p = -4; p <= 4; p++) {
	  a = m + n + p + l;
	  if (abs(a) <= 4) {
	    cmplx = CMULT(V1freqplus[m + 4], V1freqplus[n + 4]);
	    if (m != n)
	      cmplx = CMULT(two, cmplx);
	    cmplx = CMULT(cmplx, V1freqminus[p + 4]);
	    cmplx = CMULT(cmplx, IM3term[a + 4][l + 12]);
	    IM3SC[a + 4] = CADD(IM3SC[a + 4], cmplx);
	  }
	}			/*p */
      }				/*n */
    }				/*m */
  }				/*l */
  for (a = 0; a <= 8; a++) {
    IM3SC[a] = CMULT(quarter, IM3SC[a]);
    IM3SC[a] = CADD(IM23SC[a], IM3SC[a]);
  }

  /*Multiply the row of GZmtx corresponding to the output by IM3SC
     to get the IM output current, IdIM3 */
  IdIM3 = zero;
  for (b = 0; b <= 8; b++) {
    cmplx = CMULT(GZmtx[outfreq + 4][b], IM3SC[b]);
    IdIM3 = CADD(cmplx, IdIM3);
  }

  /*Finally, calculate IM2 powers */
  COM_TO_POLAR(IdIM3, &IdIMmag, &phase);
  *IM3_power = 0.5 * IdIMmag * IdIMmag * outzemb.x;	/*2*w1-w2 term */
_L625:;
}
